# -*- coding: utf-8 -*-
"""Demonstration / usage example for the new features.

Created on Fri Mar 24 13:58:36 2017

@author: Juha Jeronen, juha.jeronen@tut.fi
"""

from __future__ import division, print_function, absolute_import

import numpy as np

import matplotlib.pyplot as plt
import matplotlib.cm
import matplotlib.colors

import bspline
import bspline.splinelab as splinelab


def test():
    ########################
    # config
    ########################

    p      = 3  # order of spline basis (as-is! 3 = cubic)

    nknots = 5  # for testing: number of knots to generate (here endpoints count only once)

    tau = [0.1, 0.33]  # collocation sites (i.e. where to evaluate)

    ########################
    # usage example
    ########################

    knots = np.linspace(0,1,nknots)     # create a knot vector without endpoint repeats
    k     = splinelab.augknt(knots, p)  # add endpoint repeats as appropriate for spline order p
    B     = bspline.Bspline(k, p)       # create spline basis of order p on knots k

    # build some collocation matrices:
    #
    A0 = B.collmat(tau)                 # function value at sites tau
    A2 = B.collmat(tau, deriv_order=2)  # second derivative at sites tau

    print( A0 )
    print( A2 )

    ########################
    # tests
    ########################

    # number of basis functions
    n_interior_knots = len(knots) - 2
    n_basis_functions_expected = n_interior_knots + (p + 1)
    n_basis_functions_actual = len(B(0.))  # perform dummy evaluation to get number of basis functions
    assert n_basis_functions_actual == n_basis_functions_expected, "something went wrong, number of basis functions is incorrect"

    # partition-of-unity property of the spline basis
    assert np.allclose( np.sum(A0, axis=1), 1.0 ), "something went wrong, the basis functions do not form a partition of unity"


def main():
    """Demonstration: plot a B-spline basis and its first three derivatives."""

    #########################################################################################
    # config
    #########################################################################################

    # Order of spline basis.
    #
    p = 3

    # Knot vector, including the endpoints.
    #
    # For convenience, endpoints are specified only once, regardless of the value of p.
    #
    # Duplicate knots *in the interior* are allowed, with the standard meaning for B-splines.
    #
    knots = [0., 0.25, 0.5, 0.75, 1.]

    # How many plotting points to use on each subinterval [knots[i], knots[i+1]).
    #
    # Only intervals with length > 0 are actually plotted.
    #
    nt_per_interval = 101

    #########################################################################################
    # the demo itself
    #########################################################################################

    # The evaluation algorithm used in bspline.py uses half-open intervals  t_i <= x < t_{i+1}.
    #
    # This causes the right endpoint of each interval to actually be the start point of the next interval.
    #
    # Especially, the right endpoint of the last interval is the start point of the next (nonexistent) interval,
    # so the basis will return a value of zero there.
    #
    # We work around this by using a small epsilon to avoid evaluation exactly at t_{i+1} (for each interval).
    #
    epsrel = 1e-10
    epsabs = epsrel * (knots[-1] - knots[0])

    original_knots = knots
    knots = splinelab.augknt( knots, p )  # add repeated endpoint knots for splines of order p

    # treat each interval separately to preserve discontinuities
    #
    # (useful especially when plotting the highest-order nonzero derivative)
    #
    B   = bspline.Bspline(knots, p)
    xxs = []
    for I in zip( knots[:-1], knots[1:] ):
        t_i   = I[0]
        t_ip1 = I[1] - epsabs
        if t_ip1 - t_i > 0.:  # accept only intervals of length > 0 (to skip higher-multiplicity knots in the interior)
            xxs.append( np.linspace(t_i, t_ip1, nt_per_interval) )

    # common settings for all plotted lines
    settings = { "linestyle" : 'solid',
                 "linewidth" : 1.0 }

    # create a list of unique colors for plotting
    #
    # http://stackoverflow.com/questions/8389636/creating-over-20-unique-legend-colors-using-matplotlib
    #
    NUM_COLORS = nbasis = len( B(0.) )  # perform dummy evaluation to get number of basis functions
    cm         = plt.get_cmap('gist_rainbow')
    cNorm      = matplotlib.colors.Normalize(vmin=0, vmax=NUM_COLORS-1)
    scalarMap  = matplotlib.cm.ScalarMappable(norm=cNorm, cmap=cm)
    colors     = [scalarMap.to_rgba(i) for i in range(NUM_COLORS)]

    labels = [ r"$B$",
               r"$\mathrm{d}B\,/\,\mathrm{d}x$",
               r"$\mathrm{d}^2B\,/\,\mathrm{d}x^2$",
               r"$\mathrm{d}^3B\,/\,\mathrm{d}x^3$" ]

    # for plotting the knot positions:
    unique_knots_xx = np.unique(original_knots)
    unique_knots_yy = np.zeros_like(unique_knots_xx)

    # plot the basis functions B(x) and their first three derivatives
    plt.figure(1)
    plt.clf()
    for k in range(4):
        ax = plt.subplot(2,2, k+1)

        # place the axis label where it fits
        if k % 2 == 0:
            ax.yaxis.set_label_position("left")
        else:
            ax.yaxis.set_label_position("right")

        # plot the kth derivative; each basis function gets a unique color
        f = B.diff(order=k)  # order=0 is a passthrough
        for xx in xxs:
            yy = np.array( [f(x) for x in xx] )  # f(scalar) -> rank-1 array, one element per basis function
            for i in range(nbasis):
                settings["color"] = colors[i]
                plt.plot( xx, yy[:,i], **settings )
            plt.ylabel( labels[k] )

        # show knot positions
        plt.plot( unique_knots_xx, unique_knots_yy, "kx" )

    plt.suptitle(r"$B$-spline basis functions, $p=%d$" % (p))


if __name__ == '__main__':
    test()
    main()
    plt.show()



# -*- coding: utf-8 -*-
#
"""setuptools-based setup.py for bspline.

Supports Python 2.7 and 3.4.

Usage as usual with setuptools:
    python setup.py build
    python setup.py sdist
    python setup.py bdist_wheel --universal  # use --universal for projects that work on both py2/py3 as-is
    python setup.py install

For details, see
    http://setuptools.readthedocs.io/en/latest/setuptools.html#command-reference
or
    python setup.py --help
    python setup.py --help-commands
    python setup.py --help bdist_wheel  # or any command
"""

from __future__ import division, print_function, absolute_import

try:
    # Python 3
    MyFileNotFoundError = FileNotFoundError
except:  # FileNotFoundError does not exist in Python 2.7
    # Python 2.7
    # - open() raises IOError
    # - remove() (not currently used here) raises OSError
    MyFileNotFoundError = (IOError, OSError)

#########################################################
# General config
#########################################################

# Name of the top-level package of your library.
#
# This is also the top level of its source tree, relative to the top-level project directory setup.py resides in.
#
libname="bspline"

# Short description for package list on PyPI
#
SHORTDESC="Compute B-spline basis functions via Cox - de Boor algorithm."

# Long description for package homepage on PyPI
#
DESC="""Python/Numpy implementation of Bspline basis functions via Cox - de Boor algorithm.

Also provided are higher-order differentiation, collocation matrix generation, and a minimal procedural API (mainly for dealing with knot vectors) which may help in converting MATLAB codes.

Supports Python 2.7 and 3.4.
"""

# Set up data files for packaging.
#
# Directories (relative to the top-level directory where setup.py resides) in which to look for data files.
datadirs  = ("test",)

# File extensions to be considered as data files. (Literal, no wildcards.)
dataexts  = (".py", ".ipynb",  ".sh",  ".lyx", ".tex", ".txt", ".pdf")

# Standard documentation to detect (and package if it exists).
#
standard_docs     = ["README", "LICENSE", "TODO", "CHANGELOG", "AUTHORS"]  # just the basename without file extension
standard_doc_exts = [".md", ".rst", ".txt", ""]  # commonly .md for GitHub projects, but other projects may use .rst or .txt (or even blank).


#########################################################
# Init
#########################################################

# check for Python 2.7 or later
# http://stackoverflow.com/questions/19534896/enforcing-python-version-in-setup-py
import sys
if sys.version_info < (2,7):
    sys.exit('Sorry, Python < 2.7 is not supported')

import os

from setuptools import setup


# Gather user-defined data files
#
# http://stackoverflow.com/questions/13628979/setuptools-how-to-make-package-contain-extra-data-folder-and-all-folders-inside
#
datafiles = []
getext = lambda filename: os.path.splitext(filename)[1]
for datadir in datadirs:
    datafiles.extend( [(root, [os.path.join(root, f) for f in files if getext(f) in dataexts])
                       for root, dirs, files in os.walk(datadir)] )


# Add standard documentation (README et al.), if any, to data files
#
detected_docs = []
for docname in standard_docs:
    for ext in standard_doc_exts:
        filename = "".join( (docname, ext) )  # relative to the directory in which setup.py resides
        if os.path.isfile(filename):
            detected_docs.append(filename)
datafiles.append( ('.', detected_docs) )


# Extract __version__ from the package __init__.py
# (since it's not a good idea to actually run __init__.py during the build process).
#
# http://stackoverflow.com/questions/2058802/how-can-i-get-the-version-defined-in-setup-py-setuptools-in-my-package
#
import ast
init_py_path = os.path.join(libname, '__init__.py')
version = '0.0.unknown'
try:
    with open(init_py_path) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            print( "WARNING: Version information not found in '%s', using placeholder '%s'" % (init_py_path, version), file=sys.stderr )
except MyFileNotFoundError:
    print( "WARNING: Could not find file '%s', using placeholder version information '%s'" % (init_py_path, version), file=sys.stderr )


#########################################################
# Call setup()
#########################################################

setup(
    name = "bspline",
    version = version,
    author = "John T. Foster and Juha Jeronen",
    author_email = "juha.jeronen@tut.fi",
    url = "https://github.com/johntfoster/bspline",

    description = SHORTDESC,
    long_description = DESC,

    # CHANGE THIS
    license = "MIT",

    # free-form text field; http://stackoverflow.com/questions/34994130/what-platforms-argument-to-setup-in-setup-py-does
    platforms = ["Linux"],

    # See
    #    https://pypi.python.org/pypi?%3Aaction=list_classifiers
    #
    # for the standard classifiers.
    #
    classifiers = [ "Development Status :: 4 - Beta",
                    "Environment :: Console",
                    "Intended Audience :: Developers",
                    "Intended Audience :: Science/Research",
                    "License :: OSI Approved :: MIT License",
                    "Operating System :: POSIX :: Linux",
                    "Programming Language :: Python",
                    "Programming Language :: Python :: 2",
                    "Programming Language :: Python :: 2.7",
                    "Programming Language :: Python :: 3",
                    "Programming Language :: Python :: 3.4",
                    "Topic :: Scientific/Engineering",
                    "Topic :: Scientific/Engineering :: Mathematics",
                    "Topic :: Software Development :: Libraries",
                    "Topic :: Software Development :: Libraries :: Python Modules"
                  ],

    # See
    #    http://setuptools.readthedocs.io/en/latest/setuptools.html
    #
    setup_requires = [],
    install_requires = ["numpy"],
    provides = ["bspline"],

    # keywords for PyPI (in case you upload your project)
    #
    # e.g. the keywords your project uses as topics on GitHub, minus "python" (if there)
    #
    keywords = ["spline b-spline numpy"],

    # Declare packages so that  python -m setup build  will copy .py files (especially __init__.py).
    #
    # This **does not** automatically recurse into subpackages, so they must also be declared.
    #
    packages = ["bspline"],

    zip_safe = True,  # no Cython extensions

    # Custom data files not inside a Python package
    data_files = datafiles
)




# -*- coding: utf-8 -*-
#
"""Compute B-spline basis functions via Cox - de Boor algorithm.

Submodules:
    bspline.bspline
        OO interface (class Bspline)
    bspline.splinelab
        MATLAB-style interface and helper functions.

By default, the Bspline class from bspline.bspline is imported into this namespace when this module is loaded.
"""

from __future__ import absolute_import

# This is extracted automatically by the top-level setup.py.
__version__ = '0.1.1'

# add any imports here, if you wish to bring things into the library's top-level namespace when the library is imported.
from .bspline import Bspline

